###############################################################################
# (c) Copyright 2019 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Script to create  RPM Spec files

"""
import logging
import os
import argparse
import re
import sys
from string import Template
from lbinstall.Installer import Installer
from pprint import pprint

log = logging.getLogger()
log.setLevel(logging.INFO)


# Tools to retrieve the list fo externals
#############################################################
def getExternalsList(filename):
    '''
    Gets the list of all externals needed (as from LCG 86)
    '''
    log.warning("Loading list of externals from file")
    # Cache the externals dictionary for the version...
    # Loading the metadata
    import json
    with open(filename) as f:
        data = json.load(f)

        # Looking for LCG version
        # We need a file like so:
        # {
        #  "heptools": {
        #    "version": 84
        #    "packages":[
        #      "AIDA",
        #      "Boost",
        #      "CLHEP",
        #      "COOL"
        #    ]
        #    "package_versions":
        #     [
        #      ["genratorX, "1.0"]
        #
        #    ]
        #  }
        # }

    heptools = data["heptools"]
    lcgver = heptools["version"]
    externals_list = heptools.get("packages", list())
    externals_version_list = heptools.get("package_versions", list())
    return (lcgver, externals_list, externals_version_list)



lhcb_installer = None

class VersionFinder(object):
    """ Utility class to cache the DB """
    def __init__(self, lcgVersion, cmtconfig, siteroot="/tmp/siteroot"):
        self.siteroot = siteroot
        self.lcgVersion = lcgVersion
        self.cmtconfig = cmtconfig
        self.packages = None

    def setup(self):
        global lhcb_installer
        if lhcb_installer is None:
            lhcb_installer = Installer(self.siteroot)
        if self.packages is None:
            self.packages = set([p.name for p in lhcb_installer.remoteListPackages('LCG_%s_' % self.lcgVersion)])

    def find(self, package):
        self.setup()
        extName = "^%s.*$" % buildLCGName(self.lcgVersion, package,".*", self.cmtconfig.replace("-", "_"))
        for p in self.packages:
            if re.match(extName, p):
                return p

    def check(self, package, version):
        self.setup()
        extName = buildLCGName(self.lcgVersion, package, version, self.cmtconfig.replace("-", "_"))
        return extName in self.packages

def buildLCGName(lcgVersion, package, version, cmtconfig):
    return "LCG_%s_%s_%s_%s" % (lcgVersion, package, version, cmtconfig.replace("-", "_"))

def getRequiresList(cmtconfig, filename, siteroot="/tmp/siteroot"):
    # Setting the CMTCONFIG to the requested one
    os.environ['CMTCONFIG'] = cmtconfig

    # Retrieving the list of externals from the mentioned project
    (lcgVer, externals_list, external_versions_list) = getExternalsList(filename)
    vf = VersionFinder(lcgVer, cmtconfig)

    found = []
    missing = []

    # Looking up the versions of packages specified without version first
    for package in externals_list:
        res = vf.find(package)
        if res is None:
            missing.append(package)
            log.warning("Could not find package %s in LCG %s %s" % (package, lcgVer, cmtconfig))
        else:
            found.append(res)

    # Checking if the packages specified with version are correct
    for (package, version) in external_versions_list:
        if not vf.check(package, version):
            missing.append(package)
            log.warning("Could not find package %s %s in LCG %s %s" % (package, version, lcgVer, cmtconfig))
        else:
            found.append(buildLCGName(lcgVer, package, version, cmtconfig))

    return(lcgVer, found)


###############################################################################
# Class to build the SPEC itself
###############################################################################

class LHCbLCGMetaSpec(object):
    """ Class presenting the whole spec """

    def __init__(self, name, version, platform, requiresList, rpmroot, release="1"):
        """ Initialize with the list of RPMs """
        self.name = name
        self.rpmname = name
        self.version = version
        self.platform = platform
        self.requiresList = requiresList
        self.rpmroot = rpmroot
        self.release = release

        # Building the build dir paths
        myroot = rpmroot
        self.topdir = "%s/rpmbuild" % myroot
        self.tmpdir = "%s/tmpbuild" % myroot
        self.rpmtmp = "%s/tmp" % myroot
        self.srcdir = os.path.join(self.topdir, "SOURCES")
        self.rpmsdir = os.path.join(self.topdir, "RPMS")
        self.srpmsdir = os.path.join(self.topdir, "SRPMS")
        self.builddir = os.path.join(self.topdir, "BUILD")

        # And creating them if needed
        for d in [self.srcdir, self.rpmsdir, self.srpmsdir, self.builddir]:
            if not os.path.exists(d):
                os.makedirs(d)

        self.buildroot = os.path.join(self.tmpdir, "%s-%s-%s-buildroot" % \
                                      (self.name, self.version, self.platform))
        if not os.path.exists(self.buildroot):
            os.makedirs(self.buildroot)

    def getHeader(self):
        """ Build the SPEC Header """
        rpm_header = Template("""
%define version $version
%define platform $platform
%define platformFixed $platformFixed

%define _topdir $topdir
%define tmpdir $tmpdir
%define _tmppath $rpmtmp
%define debug_package %{nil}
%global __os_install_post /usr/lib/rpm/check-buildroot

Name: $rpmname
Version: 1.0.0
Release: $release
Vendor: LHCb
Summary: LCG  %{version} %{platform} for LHCb
License: GPL
Group: LCG
BuildRoot: %{tmpdir}/LCGLHCb-%{version}-%{platform}-buildroot
BuildArch: noarch
AutoReqProv: no
Prefix: /opt/LHCbSoft
Provides: /bin/sh
Provides: LCG_%{version}LHCb_%{platformFixed}

""").substitute(rpmname=self.rpmname,
                version=self.version,
                platform=self.platform,
                platformFixed=self.platform.replace("-", "_"),
                topdir=self.topdir,
                tmpdir=self.tmpdir,
                rpmtmp=self.rpmtmp,
                release=self.release)
        return rpm_header

    # RPM requirements for the whole package
    #############################################################

    def getRequires(self):
        rpm_requires = ""
        for r in self.requiresList:
            rpm_requires += "Requires: %s\n" % r

        return rpm_requires

    # RPM Description section
    #############################################################

    def getDescriptions(self):
        rpm_desc = """
%description
LCG externals %{version} %{platform} for LHCb

"""
        return rpm_desc

    # RPM Common section with build
    #############################################################

    def getCommon(self):
        rpm_common = """
%prep

%build

%install

%files

%post

%postun

%clean
"""
        return rpm_common

    # RPM Trailer
    #############################################################

    def getTrailer(self):
        rpm_trailer = """
%define date    %(echo `LC_ALL=\"C\" date +\"%a %b %d %Y\"`)

%changelog

* %{date} User <ben.couturier..rcern.ch>
- first Version
"""
        return rpm_trailer

    # Get the whole spec...
    #############################################################

    def getSpec(self):
        """ Concatenate all the fragments """

        rpm_global = self.getHeader() \
                     + self.getRequires() \
                     + self.getDescriptions() \
                     + self.getCommon() \
                     + self.getTrailer()

        return rpm_global


def createSpec(platform, filename, output=None, release = "1", rpmname=None, rpmroot="/tmp"):

    # Getting the list of RPMs to add
    (lcgVer, requiresList) = getRequiresList(platform, filename)

    print("LGC Version: %s" % lcgVer)
    print("Requires list:")
    pprint(requiresList)

    if rpmname is None:
        rpmname = "LCG_%sLHCb_%s" % (lcgVer, platform.replace("-", "_"))
    spec = LHCbLCGMetaSpec(rpmname, lcgVer, platform, requiresList, rpmroot, release)

    if output:
        with open(output, "w") as outputfile:
            outputfile.write(spec.getSpec())
        log.info("written output to %s" % output)
        return output
    else:
        print(spec.getSpec())
    log.info("Spec file generated")



###############################################################################
# Main method
###############################################################################
def main():
    # Parsing options
    desc = "Prepare the SPEC file the LHCb Meta RPM given the list of needed externals"
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('platform', help="Platform e.g. x86_64-centos7-gcc7-opt to compile against")
    parser.add_argument('external_list_json', help="JSON file describing the list of externals to install")
    parser.add_argument('-d', '--debug', action='store_const',
                        const=logging.DEBUG, dest='log_level')
    parser.set_defaults(log_level=logging.WARNING)
    parser.add_argument('-b', '--buildroot',
                      dest="buildroot",
                      default="/tmp",
                      action="store",
                      help="Force build root")
    parser.add_argument('-o', '--output',
                      dest="output",
                      default=None,
                      action="store",
                      help="File name for the generated specfile [default output to stdout]")
    parser.add_argument("-s", "--siteroot",
                      help="temporary directory where the RPMs will be installed before repackaging"
                           "[default: %default]",
                      default="/tmp/siteroot")
    parser.add_argument("-r", "--release",
                      help="Release number"
                           "[default: %default]",
                      default="1")
    parser.add_argument("-n", "--name",
                      help="Name of the package being produced"
                           "[default: %default]",
                      default=None)
    args = parser.parse_args()

    logging.basicConfig(level=args.log_level, stream=sys.stderr)
    createSpec(args.platform, args.external_list_json, args.output, release=args.release, rpmname=args.name, rpmroot=args.buildroot)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Helper script to update the content of the Software Database during the
# deployment of a project.
#


import os, sys, re
from LbCommon.Script import Script


# @class Check Released configs
# Main script class for to release RPMs to the repository.
class CheckReleasedConfig(Script):

    def __init__(self):
        Script.__init__(self, usage="\n\t%prog [options] releasedir",
                        description="Script update the SDB with the platforms effectively built")

    def defineOpts(self):
        """ User options """
        self.parser.add_option("-i", "--interactive", action="store_true", default=False,
                               help="Prompt before adding the platform")
        self.parser.add_option("--platform-regex", action="store", default=None,
                               help="Regexp for the platform to update")
        self.parser.add_option("--name-regex", action="store", default=None,
                               help="Regexp for the projects to update")
        self.parser.add_option("-u", "--update", action="store_true", default=False,
                               help="Really update the platforms, in dry-run mode otherwise")

    def getBuildConfig(self, builddir):
        """ Get the list of platforms built from slot-config.json """
        configfile = os.path.join(builddir, "slot-config.json")
        if not os.path.exists(configfile):
            raise Exception("%s does not exist!. Cannot load list of platforms" % configfile)

        self.log.info("Logging configurations from %s" % configfile)
        import json
        config = None
        with open(configfile) as f:
            data = json.load(f)
            config = data
        return config

    def updateSdb(self, builddir, updatemode, platformre, namere):
        """ Update the platform in the SDB """

        # Cheking that our files are there...
        builddir = os.path.abspath(builddir)
        self.log.warning("Build dir: %s" % builddir)

        if not os.path.exists(builddir):
            raise Exception("The build directory %s does not exist" % builddir)

        # Loading the build config
        config = self.getBuildConfig(builddir)

        # Now looking for binary projects...
        projname = lambda p: (p["name"].upper(), p["version"])
        projlist = []
        indepprojlist = []
        for p in config.get("projects", []):
            if p.get("platform_independent", False) and p["platform_independent"]:
                self.log.info("%s %s is platform independent" % projname(p))
                indepprojlist.append(projname(p))
            else:
                projlist.append(projname(p))

        # And finding the platforms and therefore listing the RPMs we should find...
        sanitize = lambda s: s.replace("-", "_")
        platforms = config["platforms"]

        # Dealing with binary dependent projects
        for p, v in projlist:

            if namere and not re.match(namere, p):
                self.log.warning("Ignoring %s %s due to regex" % (p, v))
                continue

            platform_found = []
            platform_missing = []
            for platform in platforms:
                prefix = "_".join([p, v, sanitize(platform)])
                # foundrpm = [ r for r in os.listdir(builddir) if r.startswith(prefix) ]
                foundrpm = [f for _, _, files in os.walk(builddir) for f in files if f.startswith(prefix)]
                if len(foundrpm) == 0:
                    platform_missing.append(platform)
                elif len(foundrpm) >= 1:
                    platform_found.append(platform)

            if len(platform_found) > 0:
                self.log.warning("%s %s: found   builds for %s" % (p, v, ", ".join(platform_found)))
            else:
                self.log.error("No platform builds found")

            if len(platform_missing) > 0:
                self.log.error("%s %s: missing builds for %s" % (p, v, ", ".join(platform_missing)))
            else:
                self.log.info("No platforms missing")

            # Now updating the sdb, calling the scripts not the python module to make sure we have the credentials
            if updatemode:
                import subprocess
                for c in platform_found:
                    if platformre and not re.match(platformre, c):
                        self.log.warning("Ignoring platform %s due to regex" % c)
                        continue
                    self.log.warning("Adding platform to DB: %s %s %s" % (p, v, c))
                    subprocess.check_call(["lb-sdb-add-platform", p, v, c])

                self.log.warning("Marking %s %s as released" % (p, v))
                subprocess.check_call(["lb-sdb-release", "-r", p, v])
            else:
                for c in platform_found:
                    if platformre and not re.match(platformre, c):
                        self.log.warning("Ignoring platform %s due to regex" % c)
                        continue
                    self.log.warning("Will add to SDB: %s %s %s" % (p, v, c))
                self.log.warning("Will release %s %s" % (p, v))

        # Releasing platform independent projects
        for p, v in indepprojlist:

            if namere and not re.match(namere, p):
                self.log.warning("Ignoring %s %s due to regex" % (p, v))
                continue

            if updatemode:
                import subprocess
                self.log.warning("Marking %s %s as released" % (p, v))
                subprocess.check_call(["lb-sdb-release", "-r", p, v])
            else:
                self.log.warning("Will release %s %s" % (p, v))

    def main(self):
        """ Main method for the script """
        if len(self.args) != 1:
            self.parser.error('Please specify the directory with the release')

        if not self.options.update:
            self.log.warning("In dry-run mode. use --update to update the SDB")
        else:
            self.log.warning("Updating the SDB")

        # Now performing the update
        rc = 0
        try:
            updated = self.updateSdb(self.args[0],
                                     self.options.update,
                                     self.options.platform_regex,
                                     self.options.name_regex)
        except Exception, e:
            self.log.error(e)
            rc = 1
        return rc


def main():
    s = CheckReleasedConfig()
    sys.exit(s.run())


if __name__ == '__main__':
    main()



###############################################################################
# (c) Copyright 2019 CERN                                                     #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Script to generate all RPM files in one go

"""
import json
import logging
import argparse
import os
import sys
from  LbAdmin.CreateLHCbLCGMetaSpec import createSpec

log = logging.getLogger()
log.setLevel(logging.INFO)

# Utils
#############################################################
def get_LCG_version(externals_filename):
    with open(externals_filename) as f:
        data = json.load(f)
    heptools = data["heptools"]
    lcgver = heptools["version"]
    return lcgver

###############################################################################
# Main method
###############################################################################
def main():

    # Parsing options
    parser = argparse.ArgumentParser(description="Prepare all the SPEC files for the externals/platforms listed")
    parser.add_argument('name_template', help="name template for the RPM of the form: LCG_{lcg_version}LHCb_{platform}")
    parser.add_argument('platform_list_file', help="File containing the list of platforms")
    parser.add_argument('external_list_json', help="JSON file describing the list of externals to install")
    parser.add_argument('-d', '--debug', action='store_const',
                       const=logging.DEBUG, dest='log_level')
    parser.set_defaults(log_level=logging.WARNING)
    parser.add_argument('-b', '--buildroot',
                      dest="buildroot",
                      default="/tmp",
                      action="store",
                      help="Force build root")
    parser.add_argument("-r", "--release",
                      help="Release number"
                           "[default: %default]",
                      default="1")
    args = parser.parse_args()
    logging.basicConfig(level=args.log_level, stream=sys.stderr)

    # Reading some date from the files
    lcg_version = get_LCG_version(args.external_list_json)
    platforms = []
    with open(args.platform_list_file) as f:
        platforms += [ l.strip() for l in f.readlines() if l]

    specs = []
    for p in platforms:
        rpmname = args.name_template.format(lcg_version=lcg_version, platform=p)
        output = rpmname + ".spec"
        f = createSpec(p, args.external_list_json, output , release=args.release,
                       rpmname=rpmname, rpmroot=args.buildroot)
        print("Created {}".format(f))
        specs.append(f)

    print("To generate the RPMs run:")
    for s in specs:
        print("rpmbuild -bb {}".format(s))


if __name__ == '__main__':
    main()
